from django.http import StreamingHttpResponse
from collections import deque
import time

class SSEManager:
    _listeners = {}

    @classmethod
    def register_listener(cls, model_name, object_id, field_name):
        if model_name not in cls._listeners:
            cls._listeners[model_name] = {}
        if object_id not in cls._listeners[model_name]:
            cls._listeners[model_name][object_id] = {}
        if field_name not in cls._listeners[model_name][object_id]:
            cls._listeners[model_name][object_id][field_name] = deque() 
        return cls._listeners[model_name][object_id][field_name]

    @classmethod
    def notify_listeners(cls, model_name, object_id, field_name, new_value):
        listeners = cls._listeners.get(model_name, {}).get(object_id, {}).get(field_name, deque())
        listeners.append(new_value)

    @classmethod
    def stream_updates(cls, request, app_label, model_name, object_id, field_name):
        def event_stream():
            listeners = cls.register_listener(model_name, object_id, field_name)

            from django.apps import apps
            model = apps.get_model(app_label, model_name)
            try:
                instance = model.objects.get(pk=object_id)
                last_value = getattr(instance, field_name)
            except model.DoesNotExist:
                last_value = None

            if last_value is not None:
                yield f"data: {last_value}\n\n"

            try:
                while True:
                    if listeners:
                        try:
                            new_value = listeners.popleft() 
                            yield f"data: {new_value}\n\n"
                        except IndexError:
                            pass 
                    yield ":\n\n"
                    time.sleep(1)
            except GeneratorExit:
                cls._listeners[model_name][object_id][field_name].clear()
                raise

        return StreamingHttpResponse(event_stream(), content_type='text/event-stream')



from django.db.models.signals import post_save
from django.dispatch import receiver
from django.apps import apps
from .sse import SSEManager
from django.db.models import Model

# You can update the signals like so for different bindings

# @receiver(post_save, sender=HighScore)
# def high_score_update_signal(sender, instance, **kwargs):
#     if instance.is_highest:
#         # Notify all listeners about the new high score and the user
#         SSEManager.notify_listeners(sender.__name__, instance.pk, 'score', instance.score)
#         SSEManager.notify_listeners(sender.__name__, instance.pk, 'user', instance.user.username)

@receiver(post_save)
def model_update_signal(sender, instance, **kwargs):

    for field in instance._meta.fields:
        field_name = field.name
        new_value = getattr(instance, field_name, None)

        if new_value is not None:
            SSEManager.notify_listeners(sender.__name__, instance.pk, field_name, new_value)


def setup_signals(app_label, model, signal_handler):
    post_save.connect(signal_handler, sender=model)


from nicegui import ui
import requests
from .config import Config
from django.utils.decorators import sync_and_async_middleware

@sync_and_async_middleware
def bind_element_to_model(element, app_label, model_name, object_id=None, fields=None, element_id=None, 
                          property_name='value', dynamic_query=None):
    if fields is None or not isinstance(fields, list):
        return
    
    host = Config.get_host()
    api_endpoint = Config.get_api_endpoint()
    model = Config.get_model(app_label, model_name)
    
    # Use dynamic queries (e.g., find by user ID or high score) if provided
    if dynamic_query:
        instance = model.objects.filter(**dynamic_query).first()
        if instance:
            object_id = instance.pk
        else:
            return  # No instance found for the dynamic query
    
    if not object_id:
        return  # Fail gracefully if object_id is still None
    
    # Fetch initial data for all fields
    def fetch_initial_data():
        data = {}
        for field_name in fields:
            url = f'{host}{api_endpoint}/{app_label}/{model_name}/{object_id}/{field_name}'
            response = requests.get(url)
            if response.status_code == 200:
                data[field_name] = response.json().get(field_name, '')
            else:
                data[field_name] = ''
        return data

    # Update data for a specific field
    def update_data(field_name, value):
        if value is None or value == '':
            pass
        else:
            url = f'{host}{api_endpoint}/{app_label}/{model_name}/{object_id}/{field_name}/'
            requests.post(url, json={field_name: value})

    # Initialize the element with combined data from all fields
    initial_data = fetch_initial_data()
    combined_data = ', '.join([f'{field}: {initial_data[field]}' for field in fields])
    setattr(element, property_name, combined_data)

    # Listener events based on the element type
    if isinstance(element, ui.input):
        listener_event = 'update:model-value'
        element_tag = 'input'
    elif isinstance(element, ui.checkbox):
        listener_event = 'update:model-checked'
        element_tag = 'input[type="checkbox"]'
    elif isinstance(element, ui.slider):
        listener_event = 'update:model-value'
        element_tag = 'input[type="range"]'
    elif isinstance(element, ui.textarea):
        listener_event = 'update:model-value'
        element_tag = 'textarea'
    elif isinstance(element, ui.button):
        listener_event = 'click'
        element_tag = 'button'
    else:
        listener_event = f'update:model-{property_name}'
        element_tag = '*'

    # Handle frontend changes by updating the respective field in the model
    def on_frontend_change(e):
        new_value = ''.join(e.args).split(', ')
        field_values = {field: value for field, value in zip(fields, new_value)}
        for field_name, value in field_values.items():
            update_data(field_name, value)

    element.on(listener_event, on_frontend_change)
    
    element.props(f'class=model-element-class id={element_id}')

    # Set up Server-Sent Events (SSE) to update the element when any field changes
    def set_value_in_element(new_data):
        combined_data = ', '.join([f'{field}: {new_data[field]}' for field in fields])
        element.set_value(combined_data)

    for field_name in fields:
        sse_url = f'{host}{api_endpoint}/sse/{app_label}/{model_name}/{object_id}/{field_name}/'
        ui.add_body_html(f"""
            <script>
                document.addEventListener("DOMContentLoaded", function() {{
                    let eventSource = new EventSource("{sse_url}");

                    eventSource.onmessage = function(event) {{
                        const newValue = event.data;
                        const element = document.querySelector('.model-element-class {element_tag}[list="{element_id}-datalist"]');

                        if (element) {{
                            let current_value = element.value || element.textContent;
                            let new_field_data = '{field_name}: ' + newValue;

                            if (current_value.includes('{field_name}:')) {{
                                // Replace only the specific field value
                                const updated_value = current_value.replace(new RegExp('{field_name}:.*?(,|$)'), new_field_data + '$1');
                                element.value = updated_value;
                            }} else {{
                                // Append if field data doesn't exist
                                element.value += ', ' + new_field_data;
                            }}
                        }} else {{
                            console.error("Element with ID", '{element_id}', "not found in the class list.");
                        }}
                    }};

                    eventSource.onerror = function(error) {{
                        console.error("SSE connection error:", error);
                    }};

                    window.addEventListener('beforeunload', function() {{
                        if (eventSource) {{
                            eventSource.close();
                        }}
                    }});
                }});
            </script>
        """)

